#include "stdafx.h"
#include "clf_random.h"
#include "clf_primitives.h"
#include "clf_geometry.h"
#include "clf_camera.h"
#include "clf_resources.h"
#include "clf_light.h"
#include "clf_font.h"

extern HWND  hWnd;						// Window Handle
extern bool keyDown[256];			// holds true / false for each key

static int m_nFPS = 0;
static int m_nFrames = 0;											// FPS and FPS Counter
static bool g_pause=false;

cCamera g_cam;
clf_stage g_stage;

// a light source
clf_light g_light(GL_LIGHT0);

// base object type
class clf_world_object
{
public:
	static clf_vector3D gravity; // the famous g

	float m_mass;				// object mass
	float m_radius;     // bounding sphere
	clf_vector3D m_pos; // position
	clf_vector3D m_vel; // velocity

	clf_world_object(float m, float r ):m_mass(m), m_radius(r){
	}
	virtual ~clf_world_object(){
	}
	virtual void step_simulation(float fElapsed){
		m_pos = m_pos+m_vel * fElapsed  + fElapsed*fElapsed*0.5f*gravity;
		m_vel = m_vel+gravity*fElapsed;
	}
	virtual void render(){
	}
	// check for collision pure virtual function
	virtual bool collides_with(clf_world_object* pob)=0;
};
clf_vector3D clf_world_object::gravity = clf_vector3D(0.f, -9.81f, 0.f);

// our game world
class clf_world
{
	static float sim_speed; // used to control game speed, 1 is realtime <1 slow motion, >1 fast
public:
	// a list of the world objects
	std::list<clf_world_object*> m_objects;
	clf_world(){
	}
	~clf_world(){
	}
	// iterate the objects and find collisions
	void collision_detection(){
		std::list<clf_world_object*>::iterator it;
		for (it=m_objects.begin(); it!=m_objects.end(); ++it)
		{
			std::list<clf_world_object*>::iterator jt;
			jt = it;
			++jt;
			clf_world_object* iob = *it;
			for ( ; jt!=m_objects.end(); ++jt)
			{
				clf_world_object* job = *jt;
				iob->collides_with(job);
			}
		}
	}
	// calculate next frame
	void step_simulation(float fElapsed){
		std::list<clf_world_object*>::iterator it;
		fElapsed *= sim_speed;
		for (it=m_objects.begin(); it!=m_objects.end(); ++it)
		{
			// move objects
			(*it)->step_simulation(fElapsed);
		}
		// check for collisions
		collision_detection();
	}
	// draw the objects
	void render(){
		std::list<clf_world_object*>::iterator it;
		for (it=m_objects.begin(); it!=m_objects.end(); ++it)
		{
			(*it)->render();
		}
	}
	// add object to world
	void add_object(clf_world_object* pob){
		m_objects.push_back(pob);
	}
};
float clf_world::sim_speed = .2f;	// used to slow down (like slow motion) the simulation
// the game world
clf_world g_w;

// the sphere object
// enable/disable definition to see the simulation with gravity effect or not
//#define NO_GRAVITY
class clf_sphere : public clf_world_object
{
public:
	clf_sphere(float m, float r ):clf_world_object(m,r){
	}
	virtual ~clf_sphere(){
	}
	virtual void render(){
		glPushMatrix();
		glTranslatef(m_pos.x, m_pos.y, m_pos.z);
		drawSphere(m_radius);
		glPopMatrix();
	}
	virtual void step_simulation(float fElapsed){
		// here we implement trajectory without gravity
#if defined NO_GRAVITY
		m_pos = m_pos+m_vel * fElapsed;
#else
		// base class implements gravity
		clf_world_object::step_simulation(fElapsed);
#endif
	}
	// calculate the result of the collision with another sphere
	void collision_respond(clf_sphere* pob){
		clf_vector3D x(pob->m_pos - m_pos);
		x.normalize();
		clf_vector3D u1x = x * dotProduct(x, m_vel);
		clf_vector3D u1y = m_vel - u1x;
		clf_vector3D u2x = -x * dotProduct(-x, pob->m_vel);
		clf_vector3D u2y = pob->m_vel - u2x;

		float e = .90;//2;
		float tot_mass = m_mass+pob->m_mass;
		clf_vector3D v1x = ((m_mass-e*pob->m_mass)/tot_mass)*u1x + ((1+e)*pob->m_mass/tot_mass)*u2x;
		clf_vector3D v2x = ((1+e)*m_mass/tot_mass)*u1x + ((pob->m_mass-e*m_mass)/tot_mass)*u2x;

		m_vel = u1y+v1x;
		pob->m_vel = u2y+v2x;

		x *= (m_radius + pob->m_radius);
		x *= 1.0001f;
		pob->m_pos = m_pos + x;
	}
	// check if the two spheres collide
	virtual bool collides_with(clf_sphere* pob){
		clf_vector3D dist_vec(pob->m_pos - m_pos);
		float dist = dist_vec.length();
		if (dist > (m_radius + pob->m_radius)	)
			return false;
		collision_respond(pob);
		return true;
	}
	// general collision detection function
	virtual bool collides_with(clf_world_object* pob){
		clf_sphere* ob = dynamic_cast<clf_sphere*>(pob);
		if (ob)
			return collides_with(ob);
		return true;
	}
};
// the two spheres of the game
clf_sphere g_s(1.8f, 0.13f);
clf_sphere g_s1(2.2f, 0.2f);

///////////////////////////////////////////////////////////////////////
void restart_game()
{
	// position the objects
	g_s.m_pos = clf_vector3D(-2.f, 0.f, 0.f);
	g_s1.m_pos = clf_vector3D(2.f, 0.f, -2.f);
	// set the speed 
	float speed = 8.f;
	// and launch
	g_s.m_vel = clf_vector3D(speed*cos(DegreesToRadians(45)),speed*sin(DegreesToRadians(45)),0);
	g_s1.m_vel = clf_vector3D(speed*cos(DegreesToRadians(135)),speed*sin(DegreesToRadians(135)),-speed*cos(DegreesToRadians(135)));
}

// this function is called when the program starts to initialize the game
void game_init()
{
	get_font_manager().create_font_2d("Bookman2D", "Bookman Old Style", 24);

	g_cam.setup(clf_vector3D(2,10,-25),       // position viewer
							clf_vector3D(0,1,0),        // setup the point of interest (where we are looking at)
							clf_vector3D(0,1,0));       // which way is up

	glEnable (GL_DEPTH_TEST);        // enable depth test
	glEnable(GL_COLOR_MATERIAL);     // use color for drawing the object

	// create and position a light source
	g_light.set_ambient(0.0f, 0.0f, 0.0f, 1.f);
	g_light.set_diffuse(0.9f, 0.9f, 0.9f, 1.f);
	g_light.set_specular(1.f, 1.f, 1.f, 1.f);
	g_light.set_position(-300.f, 100.f, -300.f, 0.f);

	// add the objects to the world
	g_w.add_object(&g_s);
	g_w.add_object(&g_s1);

	restart_game();
}

// this function is called when the window is resized
void resize_window (int width, int height)			// reshape the window when it's moved or resized
{
	g_stage.resize(0,0,width, height);

	glMatrixMode (GL_PROJECTION);									// select the projection matrix
	glLoadIdentity ();														// reset the projection matrix

	// viewport and perspective
	g_stage.prepare(); 
}

// this function is called upon exit to clear allocated resources
void game_terminate()
{
}

// the core of the game. This is called repeatedly to update the state of the game
// fElapsed: is the elapsed time since last call. The time is in seconds (usualy a very small number)
void frame_move(float fElapsed)
{
	if (keyDown[VK_ESCAPE]) 
		DestroyWindow(hWnd);
	if (keyDown['R'])
	{
		keyDown['R'] = 0;
		restart_game();
	}
	if (keyDown['P'])
	{
		keyDown['P'] = 0;
		g_pause = !g_pause;
	}

	if (!g_pause)
	{
		g_w.step_simulation(fElapsed);
	}

	// frame counting mechanism
	// the higher the frame rate the faster the system
	// if frame rate drops below 30 we are in deep trouble
	// we should either optimize the program or buy a new computer
	static dFloat tot = 0;
	tot += fElapsed;
	m_nFrames++;												// Increment Our frame Counter
	if (tot >= 1.f)
	{
		tot = 0;
		m_nFPS = m_nFrames;
		m_nFrames = 0;

		char txt[200];
		sprintf(txt, "GusOnGames, fps:%d",m_nFPS);
		SetWindowText(hWnd, txt);
	}
}

// the drawing function
// it is called right after the frame_move function
void frame_render()
{
	// clear the screen
	glClearColor (0.f, 0.f, 0.2f, 1.f);				// background colour
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// clear screen and depth buffer

	glMatrixMode (GL_MODELVIEW);								// select the modelview matrix
	glLoadIdentity ();													// reset

	// position the viewer
	g_cam.apply();

	glBegin(GL_QUADS);
	glNormal3f(0,1,0);
	glVertex3f(-10,0,-10);
	glVertex3f(-10,0,10);
	glVertex3f(10,0,10);
	glVertex3f(10,0,-10);
	glEnd();
	
	// position the light
	g_light.enable();
	g_light.apply();

	g_w.render();

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_TEXTURE_2D);

	clf_font* gfnt = get_font_manager().get_font("Bookman2D");

	// comment or uncomment the folloing line to activate the two different positioning modes
#define FONT_USE_SCRFEEN

#ifdef FONT_USE_SCRFEEN
	gfnt->use_screen_position(true);
	gfnt->set_position_2d(10, 10);
#else
	gfnt->use_screen_position(false);
	glRasterPos2f(-2, -2);
#endif

	clf_vector3D momentum = g_s.m_vel*g_s.m_mass + g_s1.m_vel*g_s1.m_mass;
	float k1 = g_s.m_mass*dotProduct(g_s.m_vel,g_s.m_vel)/2;
	float k2 = g_s1.m_mass*dotProduct(g_s1.m_vel,g_s1.m_vel)/2;

	gfnt->set_color_RGB(1.f,1.f,1.f);
	gfnt->print_2d("fps:%d, momentum:%.3f, energy:%.3f",m_nFPS, momentum.length(), k1+k2);
	glPopAttrib();
}
